# Import the necessary libraries
import os
import binascii
import cv2
import string
import sqlite3 # Import not needed in the integrated system
import pandas as pd # Import not needed in the integrated system
import math
from difflib import SequenceMatcher
from getpass import getpass



# Define the functions

# Function to set up a device; function not needed in the integrated system; not checked
def deviceSetup():
    db = sqlite3.connect(':memory:')
    cursor = db.cursor()

    cursor.execute('''CREATE TABLE biometrics(id INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL, template BLOB, csprng INT)''')
    db.commit()
    

# Function to generate a random number; checked
def randomNumberGeneration():
    csprngVal = int(binascii.hexlify(os.urandom(3)), 16)
    print("Generated CSPRNG value: ", csprngVal) # Line not needed in the integrated system
    return csprngVal
    
    
# Function to generate the payload; checked
def payloadGeneration(csprngVal, incomingHexVal):
    decimalHexVal = int(incomingHexVal, 16)
    print("Hexadecimal value converted to decimal: ", decimalHexVal) # Line not needed in the integrated system
    payloadInt = csprngVal + decimalHexVal
    payload = str(payloadInt)
    print("Generated payload: ", payloadInt) # Line not needed in the integrated system
    return payload


# Steganography encoding function; checked
def encode(passphrase, incomingTemplate, payload):
    # Null arrays to for data injection
    d = {}
    c = {}

    for i in range(255):
        d[chr(i)] = i
        c[i] = chr(i)

    template = cv2.imread(incomingTemplate)

    # Obtain the dimensions of the image
    i = template.shape[0]
    j = template.shape[1]

    print("Dimensions of the image: ", i, " x ", j) # Line not needed in the integrated system

    keyLength = 0 # Initial value of the key length
    z = 0 # Obtain the plane
    n = 0 # Number of rows
    m = 0 # Number of columns

    payloadLength = len(payload)

    for i in range(payloadLength):
        template[n,m,z]=d[payload[i]]^d[passphrase[keyLength]]
        n += 1 # Spacing
        m += 1
        m = (m + 1) % 3 # Print values in a vertical line
        keyLength = (keyLength + 1) % len(passphrase)

    cv2.imwrite(saveTemplate, template) 
    os.startfile(saveTemplate) # Line not needed in the integrated system
    print("Data Hiding in Image completed successfully.") # Line not needed in the integrated system
    return saveTemplate
    
    
# Steganography decoding function; checked
def decode(passphrase, saveTemplate):
    template = cv2.imread(saveTemplate)
    
    keyLength = 0
    z = 0 # Obtain the plane
    n = 0 # Number of row
    m = 0 # Number of column
    
    payloadLength = len(payload)
    decryptedPayload = ""
    
    for i in range(payloadLength):
        decryptedPayload += c[template[n,m,z]^d[passphrase[keyLength]]]
        n += 1
        m += 1
        m = (m + 1) % 3
        keyLength = (keyLength + 1) % len(passphrase)
    
    return decryptedPayload


# Function to update the database; function not needed in the integrated system; not checked
def updateDatabase(saveTemplate):
    db = sqlite3.connect(':memory:')
    cursor = db.cursor()

    with open(saveTemplate, 'rb') as f:
        img = f.read()
        
    safeVal = csprngVal * math.pi

    cursor.execute('''INSERT INTO biometrics(saveTemplate, csprng) VALUES (?,?)''', (img, safeVal))
    db.commit()


# Function to retrieve data from the database; function not needed in the integrated system; not checked
def retrieveDatabase():
    db = sqlite3.connect(':memory:')
    cursor = db.cursor()
    
    selectRecords = "SELECT * FROM biometrics"
    cursor.execute(selectRecords)
    records = cursor.fetchall()
    
    return records


# Function to score the similarity between two payloads
def similarityScorer(refPayload, checkPayload):
    return SequenceMatcher(None, refPayload, checkPayload).ratio()


# Function to authenticate the user; works for one user; add loop for database if needed; checked
def authenticate(checkPayload, refPayload):
    thresholdScore = 0.889999
    similarityScore = similarityScorer(refPayload, checkPayload)
    
    if similarityScore >= thresholdScore:
        print("Authentication Successful!")
        # Direct user to data
    
    else:
        print("Unsuccessful Authentication!")
        # Check what to do with the other components


# Function for the passphrase; checked
def passphrase():
    passphrase1 = getpass(prompt = 'Please enter a one-time passphrase of your choice: ')
    passphrase2 = getpass(prompt = 'Re-enter the one-time passphrase: ')
    
    if passphrase1 == passphrase2:
        pass
        
    else:
        print("Error! Mismatching passphrases.")
        passphrase1 == passphrase2 == ''
        passphrase()
    
    return passphrase1

    
# Function for the Enrollment Phase; checked
def enrollment(): # Send the hexadecimal value (x) and the fingerprint template as inputs (y)
    passphrase = str(passphrase()) # Encrypt before storing in the database (DO NOT HASH)
    csprngVal = randomNumberGeneration() # Multiply by math.pi before storing in the database
    payload = str(payloadGeneration(csprngVal, 'a12b34c5')) # (x)
    saveTemplate = encode(passphrase, 'labrador.jpg', payload) # (y)
    # Save the passphrase (encrypted), CSPRNG value and the fingerprint template in the database.
    updateDatabase(saveTemplate) # Function not needed in the integrated system


# Function for the Authentication Phase
def authentication(): # Send the new hexadecimal value (z) as an input
    # First, obtain the data from the database (CSPRNG value, passphrase, template)
    csprngVal = retrCSPRNG / math.pi
    checkPayload = str(payloadGeneration(csprngVal, 'a12b34d6')) # (z)
    refPayload = decode(passphrase, saveTemplate)
    authenticate(checkPayload, refPayload)